"""Functions to access https://toar-data.fz-juelich.de/api/v2/"""
__author__ = 'Lukas Leufen'
__date__ = '2022-06-30'


import logging
from typing import Union, List, Dict
from io import StringIO

import pandas as pd
import pytz
from timezonefinder import TimezoneFinder

from mlair.configuration.toar_data_v2_settings import toar_data_v2_settings
from mlair.helpers import to_list
from mlair.helpers.data_sources.toar_data import EmptyQueryResult, get_data, correct_stat_name


str_or_none = Union[str, None]


def download_toar(station_name: Union[str, List[str]], stat_var: dict,
                  sampling: str = "daily", data_origin: Dict = None):
    """
    Download data from https://toar-data.fz-juelich.de/api/v2/

    Uses station name to indicate measurement site and keys of stat_var to indicate variable name. If data origin is
    given, this method tries to load time series for this origin. In case no origin is provided, this method loads data
    with the highest priority according to toar-data's order parameter.

    :param station_name:
    :param stat_var:
    :param sampling:
    :param data_origin:
    :return:
    """

    # make sure station_name parameter is a list
    station_name = to_list(station_name)

    # also ensure that given data_origin dict is no reference
    if data_origin is None or len(data_origin) == 0:
        data_origin = None
    else:
        data_origin = {k: v for (k, v) in data_origin.items()}

    # get data connection settings for meta
    meta_url_base, headers = toar_data_v2_settings("meta")

    # load variables
    var_meta = load_variables_information(stat_var, meta_url_base, headers)

    # load station meta
    station_meta = load_station_information(station_name, meta_url_base, headers)

    # load series information
    timeseries_meta = load_timeseries_information(station_meta, var_meta, meta_url_base, headers, data_origin)

    # # correct stat_var values if data is not aggregated (hourly)
    # if sampling == "hourly":
    #     stat_var = {key: "values" for key in stat_var.keys()}

    logging.info(f"load data for {station_meta['codes'][0]} from TOAR-DATA")
    # get data connection settings for data
    data_url_base, headers = toar_data_v2_settings(sampling)

    data_dict = {}
    for var, meta in timeseries_meta.items():
        logging.debug(f"load {var}")
        meta_and_opts = prepare_meta(meta, sampling, stat_var, var)
        data_var = []
        for var_meta, opts in meta_and_opts:
            data_var.extend(load_timeseries_data(var_meta, data_url_base, opts, headers, sampling))
        data_dict[var] = merge_data(*data_var, sampling=sampling)
    data = pd.DataFrame.from_dict(data_dict)
    data = correct_timezone(data, station_meta, sampling)

    meta = combine_meta_data(station_meta, {k: v[0] for k, v in timeseries_meta.items()})
    meta = pd.DataFrame.from_dict(meta, orient='index')
    meta.columns = station_name
    return data, meta


def merge_data(*args, sampling="hourly"):
    start_date = min(map(lambda x: x.index.min(), args))
    end_date = max(map(lambda x: x.index.max(), args))
    freq = {"hourly": "1H", "daily": "1d"}.get(sampling)
    full_time = pd.date_range(start_date, end_date, freq=freq)
    full_data = args[0].reindex(full_time)
    if not isinstance(full_data, pd.DataFrame):
        full_data = full_data.to_frame()
    for d in args[1:]:
        full_data.update(d, overwrite=False)
    return full_data.squeeze()


def correct_timezone(data, meta, sampling):
    """
    Extract timezone information and convert data index to this timezone.

    Uses UTC if no information is provided. Note that is method only modifies data in with sampling='hourly'. In all
    other cases, it returns just the given data without any change. This method expects date index of data to be in UTC.
    Timezone information is not added to the index to get rid of daylight saving time and ambiguous timestamps.
    """
    if sampling == "hourly":
        tz_info = meta.get("timezone", "UTC")
        try:
            tz = pytz.timezone(tz_info)
        except pytz.exceptions.UnknownTimeZoneError as e:
            lon, lat = meta["coordinates"]["lng"], meta["coordinates"]["lat"]
            tz = pytz.timezone(TimezoneFinder().timezone_at(lng=lon, lat=lat))
        index = data.index
        index = index.tz_localize(None)
        utc_offset = tz.utcoffset(index[0]) - tz.dst(index[0])
        data.index = index + utc_offset
    return data


def prepare_meta(meta, sampling, stat_var, var):
    out = []
    for m in meta:
        opts = {}
        if sampling == "daily":
            opts["timeseries_id"] = m.pop("id")
            m["id"] = None
            opts["names"] = stat_var[var]
            opts["sampling"] = sampling
        out.append(([m], opts))
    return out


def combine_meta_data(station_meta, timeseries_meta):
    meta = {}
    for k, v in station_meta.items():
        if k == "codes":
            meta[k] = v[0]
        elif k in ["coordinates", "additional_metadata", "globalmeta"]:
            for _key, _val in v.items():
                if _key == "lng":
                    meta["lon"] = _val
                else:
                    meta[_key] = _val
        elif k in ["changelog", "roles", "annotations", "aux_images", "aux_docs", "aux_urls"]:
            continue
        else:
            meta[k] = v
    for var, var_meta in timeseries_meta.items():
        for k, v in var_meta.items():
            if k in ["additional_metadata", "station", "programme", "annotations", "changelog"]:
                continue
            elif k == "roles":
                for _key, _val in v[0]["contact"]["organisation"].items():
                    new_k = f"{var}_organisation_{_key}"
                    meta[new_k] = _val
            elif k == "variable":
                for _key, _val in v.items():
                    new_k = f"{var}_{_key}"
                    meta[new_k] = _val
            else:
                new_k = f"{var}_{k}"
                meta[new_k] = v
    return meta


def load_timeseries_data(timeseries_meta, url_base, opts, headers, sampling):
    coll = []
    for meta in timeseries_meta:
        series_id = meta["id"]
        # opts = {"base": url_base, "service": f"data/timeseries/{series_id}"}
        opts = {"base": url_base, "service": f"data/timeseries", "param_id": series_id, "format": "csv", **opts}
        if sampling != "hourly":
            opts["service"] = None
        res = get_data(opts, headers, as_json=False)
        data = pd.read_csv(StringIO(res), comment="#", index_col="datetime", parse_dates=True,
                           infer_datetime_format=True)
        if len(data.index) > 0:
            data = data[correct_stat_name(opts.get("names", "value"))].rename(meta["variable"]["name"])
            coll.append(data)
    return coll


def load_station_information(station_name: List[str], url_base: str, headers: Dict):
    # opts = {"base": url_base, "service": f"stationmeta/{station_name[0]}"}
    opts = {"base": url_base, "service": f"stationmeta", "param_id": station_name[0]}
    return get_data(opts, headers)


def load_timeseries_information(station_meta, var_meta, url_base: str, headers: Dict,
                                data_origin: Dict = None) -> [Dict, Dict]:
    timeseries_id_dict = {}
    missing = []
    for var, meta in var_meta.items():
        timeseries_id_dict[var] = []
        opts = {"base": url_base, "service": "search", "station_id": station_meta["id"], "variable_id": meta["id"]}
        res = get_data(opts, headers)
        if len(res) == 0:
            missing.append((var, meta))
            # raise EmptyQueryResult(f"Cannot find any timeseries for station id {station_meta['id']} "
            #                        f"({station_meta['codes'][0]}) and variable id {meta['id']} ({var}).")
        if data_origin is not None:
            var_origin = data_origin[var]
            timeseries_id_dict[var] = select_timeseries_by_origin(res, var_origin)
            # if len(timeseries_id_dict[var]) == 0:
            #     raise EmptyQueryResult(f"Cannot find any timeseries for station id {station_meta['id']} "
            #                            f"({station_meta['codes'][0]}), variable id {meta['id']} ({var}) "
            #                            f"and timeseries origin {var_origin}.")
        if data_origin is None or len(timeseries_id_dict[var]) == 0:
            timeseries_id_dict[var] = select_timeseries_by_order(res)
    if len(missing) > 0:
        missing = ",".join([f"{m[0]} ({m[1]['id']})" for m in missing])
        raise EmptyQueryResult(f"Cannot find any timeseries for station id {station_meta['id']} "
                               f"({station_meta['codes'][0]}) and variables {missing}.")
    return timeseries_id_dict


def select_timeseries_by_order(toar_meta):
    order_dict = {meta["order"]: meta for meta in toar_meta}
    res = [order_dict[order] for order in sorted(order_dict.keys())]
    return res


def select_timeseries_by_origin(toar_meta, var_origin):
    res = []
    for origin in to_list(var_origin):
        for meta in toar_meta:
            for roles in meta["roles"]:
                if roles["contact"]["organisation"]["name"].lower() == origin.lower():
                    res.append(meta)
                    break
    return res


def load_variables_information(var_dict, url_base, headers):
    var_meta_dict = {}
    for var in var_dict.keys():
        opts = {"base": url_base, "service": f"variables", "param_id": var}
        var_meta_dict[var] = get_data(opts, headers)
    return var_meta_dict
